# $Date: 1996/12/09 16:43:20 $ $Author: olli $ $Revision: 1.3.2.2 $ #

# kg, 28/07/95 #

#++
plotfunc -- plot function graphs

plotfunc(f1,..., fn, x=xl..xu [, y=yl..yu])
plotfunc(f1,..., fn, x=[xl,xu] [, y=[yl,yu]])

f1,...,fn      - expression
x, y           - identifiers
xl, xu, yl, yu - real numbers

If only one equation x=xl..xu is given, then the fi are interpreted as
functions in x from the interval [xl,xu] into the real numbers.

If two equations x=xl..xu and y=yl..yu are given, the fi are interpreted as
functions in x and y from the range [xl,xu] x [yl,yu] into the real
numbers.

The graphs of these functions are then plotted.

Note that the fi are not evaluated, but the equations only.

Examples:
plotfunc(sin(x), cos(x), x=-[PI,PI]);	-- draws sin and cos on [-PI,PI]
plotfunc(sin(x*y)^2, x=0..1, y=0..1);	-- draws x,y -> sin(x*y)^2 on [0,1]x[0,1]
++#

stdlib::RATIO_YX := 10:

plotfunc:= proc()
    option hold;
    local a, i, m, p, f, xr, yr, DIGITS, getrange, t,
          ac, l, x, c, d, grid, iv;
begin
    DIGITS:= 10;
    if args(0) < 2 then error("wrong no of args") end_if;
    
    getrange:= proc(r) begin
        if domtype(op(r,1)) <> DOM_IDENT then error("no variable") end_if;
        if type(op(r,2)) = "_range" then
	    r:= (op(r,1) = [op(op(r,2))])
	elif domtype(op(r,2)) <> DOM_LIST then
	    error("no plot range")
	elif nops(op(r,2)) <> 2 then
	    error("no plot range")
	end_if;
	r:= (op(r,1) = map(op(r,2), float));
	if domtype(op(r, [2,1])) <> DOM_FLOAT then 
	    error("illegal plot range") 
	end_if;
	if domtype(op(r, [2,2])) <> DOM_FLOAT then 
	    error("illegal plot range") 
	end_if;
        r
    end_proc;
    
    # get ranges #
    i:= args(0);
    a:= args(i);
    if type(a) <> "_equal" then error("no plot range") end_if;
    yr:= getrange(context(a));
    i:= i-1;
    a:= args(i);
    if type(a) = "_equal" then
        m:= hold(Mode=Surface);
        xr:= getrange(context(a));
        i:= i-1;
        if i = 0 then error("no function to plot") end_if;
    else
        m:= hold(Mode=Curve);
        xr:= yr;
        yr:= FAIL;
    end_if;
    
    # get functions #
    f:= null();
    t:= "";
    while i > 0 do
        # allow for plotting functions #
        a:= context(args(i)); ac := a;
        if contains({DOM_PROC, DOM_EXEC, DOM_FUNC_ENV}, domtype(a)) then
            if yr = FAIL then
		a:= subsop(hold(f(x)), 0=a, 1=op(xr,1))
	    else
	        a:= subsop(hold(f(x,y)), 0=a, 1=op(xr,1), 2=op(yr,1))
	    end_if
        else
            a:= args(i)
        end_if;
	# add graph definition #
	if yr = FAIL then

        #------------ check for singularities ---------------#

        grid := 100;
        x := op(xr,1); c := op(xr, [2,1]); d := op(xr, [2,2]);
        l := stdlib::get_intervals(ac, x, c, d, stdlib::RATIO_YX, grid );
        for iv in l do
	        f:= f, [m, [ x, a ], x=iv, Grid=[grid]]
        end_for
	else
	    f:= f, [m, [ op(xr,1), op(yr,1), a ], xr, yr]
	end_if;
	if t = "" then
	    t:= expr2text(args(i))
	else
	    t:= expr2text(args(i)).", ".t
	end_if;
	i:= i-1;
    end_while;
    
    # define scene attributes #
    if yr = FAIL then 
	p:= plot2d;
	xr:= expr2text(op(xr,1));
	a:= (Scaling=UnConstrained, Axes=Corner, Labeling=TRUE,
	     Labels=[xr, ""], Title=t)
    else
	p:= plot3d;
	xr:= expr2text(op(xr,1));
	yr:= expr2text(op(yr,1));
	a:= (Scaling=UnConstrained, Labeling=TRUE,
	     Labels=[xr, yr, ""], Title=t)
    end_if;
    
    # plot the stuff #
    p(a, f);
end_proc:

# end of 'plotfunc' #


#--

Diese Prozedur untersucht, ob die Funktion f(x) im endlichen
Intervall [a,b] Singularitaeten besitzt und teilt ggf. die Intervalle
auf.

Enthaelt f an der Stelle d eine Singularitaet, so wird ein
maximaler bzw. minimaler Wert f"ur die Funktion gesetzt, ein sogenannter
threshold. 

Ergebnis hat folgende Form: [[a_1,b_1], ... , [a_n, b_n]], n >= 1.

Bsp: 

>> get_intervals(1/cos(x),x,0,PI,10,100);

             [[0.0, 1.507091268], [1.634501384, 3.141592653]]

>> get_intervals(10*x/(x^2-2*x-8),x,-5,5,10,100);

  [[-5.0, -2.072061831], [-1.938005503, 3.860888646], [4.127955252, 5.0]]

Ohne Verfeinerung mittels Bisektion w"urden die folgenden Intervalle geliefert:

                [[-5.0, -2.09], [-1.88, 3.82], [4.13, 5.0]]

--#


stdlib::get_intervals := proc(f, x, a, b, ratio, grid) 
   local ds, f1, i, l, n, t,
         max_y, min_y, new_bound, bisection,
         eps, thres_max_y, thres_min_y;
begin

   bisection := proc(a, b, bound)
      local c;
   begin
      eps := 10^(-DIGITS);
      while abs(b-a) > eps do
         c := (a+b)/2;
         if f(c) <= bound then
            a := c;
         else
            b := c;
         end_if
      end_while;
      c;
   end_proc;

   # Berechne a, sodass thres_min_y <= f(a) <= thres_max_y #

   new_bound := proc(a, eps)
      local c, y;
   begin
      c := 0;
      repeat
        a := a + eps; y := f(a); c := c + 1;
      until thres_min_y <= y and y <= thres_max_y end_repeat;

      # Der folgende Abschnitt dient nur zur Verfeinerung des Wertes 'a' #
      # Diese verfeinert insbes. die graphische Ausgabe.                 #
      # Vor Ausf"uhrung des Abschnitts gilt entweder                     #
      # 1) f(a) <= thres_max_y und f(a-eps) > thres_max_y , oder         #
      # 2) f(a) >= thres_min_y und f(a-eps) < thres_min_y                #
      # Vor Ausf"uhrung des Abschnitts gilt entweder                     #
      # 1) f(a) = thres_max_y oder 2) f(a) = thres_min_y                 #

      if c > 1 then
         y := f(a-eps);
         if y > thres_max_y then
            a := bisection(a, a-eps, thres_max_y)
         else
            a := bisection(a-eps, a, thres_min_y)
         end_if
      end_if;

      a

   end_proc;

   #------------- Hauptprogramm 'get_intervals' ----------------------#

   case domtype(f)

     of DOM_EXEC do # f = func(..., x) #
        t := misc::subsFunArgs(f);
        if t = FAIL then
           return([[a,b]])
        end_if;
        f1 := t[1]; x := t[2][1];
        break;

     of DOM_EXPR do
        f1 := f;
        f :=  newpurefunc("fun(...)", NIL, 
                  misc::subsFreeVars(f, x=hold(args(1))),
                  FALSE);
        break;
   
     otherwise           # Prozeduren koennen z.B. #
        return([[a,b]])  # nicht behandelt werden. #

   end_case;

   ds := discont(f1, x=a..b);

   # Enthaelt f keine Singularitaeten oder konnten keine #
   # Singularitaeten ermittelt werden, so gebe Funktion  #
   # f und das Intervall [a, b] unveraendert zurueck.    #

   if type(ds) <> DOM_SET or ds = {} then 
      return([[a,b]])
   end_if:

   min_y := infinity; max_y := -infinity;

   ds := sort(map([op(ds)], float));
   a := float(a); b := float(b);
   n := nops(ds); 

   if ds[1] <> a then
      t := f(a);
      min_y := min(min_y, t);                                 
      max_y := max(max_y, t); 
   end_if;
   if ds[n] <> b then                                         
      t := f(b);                              
      min_y := min(min_y, t);                                 
      max_y := max(max_y, t);                                 
   end_if; 
   
   if n > 1 then
      t := f(ds[i-1]+(ds[i]-ds[i-1])/2) $ i=2..n;
      min_y := min(min_y, t); 
      max_y := max(max_y, t);
   end_if;

   t := max(0, ( ratio * (b-a) - (max_y - min_y) ) / 2);
   thres_max_y := max_y + t; thres_min_y := min_y - t;

    if ds[1] = a then                                          
      l := [];                                                
   else                                                       
      eps := (ds[1]-a)/grid;
      l := [[a, new_bound(ds[1], -eps)]];                                  
   end_if;                                                    
   for i from 2 to n do                                       
       eps := (ds[i]-ds[i-1])/grid;
       l := append(l, [new_bound(ds[i-1], +eps), new_bound(ds[i], -eps)])               
   end_for;                                                   
   if ds[n] <> b then                                         
      eps := (b-ds[n])/grid;
      l := append(l, [new_bound(ds[n], +eps), b]);                         
   end_if;

   l;

end_proc:

# end of file #
